#ifndef BatchLab_Function_h
#define BatchLab_Function_h

// Inheritance :
#include <BatchLab/Core/BaseObject.h>
#include <AIDA/IFunction.h>
#include <Slash/Meta/IObject.h>

#include <BatchLab/Core/Annotation.h>

#include <Lib/Functions.h>

namespace Slash {namespace Data {class IFunction;}}

// For compiled :
namespace Slash {namespace Data {class IArray;}}

namespace BatchLab {

class BaseFunction 
:public BaseObject
,public virtual AIDA::IFunction 
//FIXME ,public virtual AIDA::IModelFunction 
,public virtual Slash::Store::IStorable {
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
  virtual std::string type() const; //AIDA-v3r3p0
public: //AIDA/IFunction
  virtual std::string title() const;
  virtual bool setTitle(const std::string&);
  virtual AIDA::IAnnotation& annotation();
  virtual const AIDA::IAnnotation& annotation() const;
  virtual bool isEqual(const IFunction&) const;
  virtual const std::vector<double>& gradient(const std::vector<double>&) const;  
  virtual bool providesGradient() const;
  virtual std::string variableName(int) const;
  virtual std::vector<std::string> variableNames() const;
  virtual std::vector<std::string> parameterNames() const;
  virtual const std::vector<double>& parameters() const;
  virtual int numberOfParameters() const;
  virtual int indexOfParameter(const std::string&) const;
  virtual std::string codeletString() const;
  virtual bool setParameter(const std::string&,double);
  virtual std::string normalizationParameter() const; //AIDA-v3r3p0
  virtual double value(const std::vector<double>&) const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //For OnXLab::PlottableFunction :
  virtual bool value(const std::vector<double>&,double&) const = 0;
protected:
  BaseFunction(Slash::Core::IWriter&,const std::string&,const std::string&);
  BaseFunction(const BaseFunction&);
public:
  virtual ~BaseFunction();
  Annotation& getAnnotation();
  void setAnnotation(const Annotation&);
  SLASH_STORE_HEADER
protected:
  Annotation fAnnotation;
private:
  std::string fTitle;
protected:
  std::string fCodelet;
  std::vector<double> fParams;
  std::vector<std::string> fParamNames;
  std::vector<std::string> fVariableNames;
  std::vector<double> fGradient;
};

namespace Function {

class Gaussian 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Gaussian(Slash::Core::IWriter&,const std::string& = "");
  Gaussian(Slash::Core::IWriter&,
           const std::string&,
           const std::string&,
           double = 1,double = 0,double = 1);
  Gaussian(const Gaussian&);
  virtual ~Gaussian();
  SLASH_STORE_HEADER
private:
  Lib::Function::Gaussian fFunction;
};


class BreitWigner 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  BreitWigner(Slash::Core::IWriter&,const std::string& = "");
  BreitWigner(Slash::Core::IWriter&,
              const std::string&,
              const std::string&,
              double = 1,double = 0,double = 1);
  BreitWigner(const BreitWigner&);
  virtual ~BreitWigner();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::BreitWigner fFunction;
};


class ExpHBOOK 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  ExpHBOOK(Slash::Core::IWriter&,const std::string& = "");
  ExpHBOOK(Slash::Core::IWriter&,
              const std::string&,
              const std::string&,
              double = 0,double = 1);
  ExpHBOOK(const ExpHBOOK&);
  virtual ~ExpHBOOK();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Exponential fFunction;
};


class ExpAIDA 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  ExpAIDA(Slash::Core::IWriter&,const std::string& = "");
  ExpAIDA(Slash::Core::IWriter&,
               const std::string&,
               const std::string&,
               double = 1,double = 1);
  ExpAIDA(const ExpAIDA&);
  virtual ~ExpAIDA();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Exponential fFunction;
};


class Polynomial 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Polynomial(Slash::Core::IWriter&,const std::string& = "");
  Polynomial(Slash::Core::IWriter&,
             const std::string&,
             const std::string&,
             const std::vector<double>&);
  Polynomial(const Polynomial&);
  virtual ~Polynomial();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Polynomial fFunction;
};

class Cauchy 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Cauchy(Slash::Core::IWriter&,const std::string& = "");
  Cauchy(Slash::Core::IWriter&,
           const std::string&,
           const std::string&,
           double = 1,double = 0,double = 1);
  Cauchy(const Cauchy&);
  virtual ~Cauchy();
  SLASH_STORE_HEADER
private:
  Lib::Function::Cauchy fFunction;
};

class Scripted 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Scripted(Slash::Core::IWriter&,const std::string& = "");
  Scripted(Slash::Core::IWriter&,
                   const std::string&,
                   const std::string&,
                   const std::string&);
  Scripted(const Scripted&);
  virtual ~Scripted();
  bool isValid() const;
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Scripted fFunction;
};

class Composite 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
  virtual bool setParameter(const std::string&,double);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Composite(Slash::Core::IWriter&,const std::string& = "");
  Composite(Slash::Core::IWriter&,
                   const std::string&,
                   const std::string&,
                   const std::string&);
  Composite(const Composite&);
  virtual ~Composite();
  bool isValid() const;
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Composite fFunction;
};

class Compiled 
:public BaseFunction
,public virtual Slash::Meta::IObject {
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual bool setParameter(const std::string&,double);
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Compiled(Slash::Core::IWriter&,const std::string& = "");
  Compiled(Slash::Core::IWriter&,
           const std::string&,
           const std::string&,
           Slash::Data::IFunction*,
           const std::vector<Slash::Data::IArray*> arrays = 
             std::vector<Slash::Data::IArray*>() );
  Compiled(const Compiled&);
  virtual ~Compiled();
  Slash::Data::IFunction* slashFunction() const;
private:
  SLASH_STORE_HEADER
private:
  void reset();
private:
  Slash::Data::IFunction* fFunction;
  std::string fModel;
  std::string fFile;
  std::string fSignature;
  std::vector<Slash::Data::IArray*> fArrays;
};

class Array 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual const std::vector<double>& parameters() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Array(Slash::Core::IWriter&,const std::string& = "");
  Array(Slash::Core::IWriter&,
                const std::string&,
                const std::string&);
  Array(Slash::Core::IWriter&,
                const std::string&,
                const std::string&,
                const std::vector<double>&,
                const std::vector<double>&);
  Array(const Array&);
  virtual ~Array();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Array fFunction;
};

class Array2D 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual const std::vector<double>& parameters() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  Array2D(Slash::Core::IWriter&,const std::string& = "");
  Array2D(Slash::Core::IWriter&,
                  const std::string&,
                  const std::string&);
  Array2D(Slash::Core::IWriter&,
                  const std::string&,
                  const std::string&,
                  const std::vector<double>&,
                  const std::vector<double>&,
                  const std::vector<double>&);
  Array2D(const Array2D&);
  virtual ~Array2D();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::Array2D fFunction;
};

class ArrayComplex 
:public BaseFunction 
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
public: //AIDA/IFunction
  virtual int dimension() const;
  virtual const std::vector<double>& parameters() const;
  virtual bool setParameters(const std::vector<double>&);
  virtual double parameter(const std::string&) const;
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //BaseFunction
  virtual bool value(const std::vector<double>&,double&) const;
public:
  ArrayComplex(Slash::Core::IWriter&,const std::string& = "");
  ArrayComplex(Slash::Core::IWriter&,
                       const std::string&,
                       const std::string&);
  ArrayComplex(Slash::Core::IWriter&,
                       const std::string&,
                       const std::string&,
                       const std::vector<double>&,
                       const std::vector< std::complex<double> >&);
  ArrayComplex(Slash::Core::IWriter&,
                       const std::string&,
                       const std::string&,
                       const std::vector<double>&,
                       const std::vector<double>&,
                       const std::vector<double>&);
  ArrayComplex(const ArrayComplex&);
  virtual ~ArrayComplex();
private:
  SLASH_STORE_HEADER
private:
  Lib::Function::ArrayComplex fFunction;
};

} //Function

} //BatchLab

#endif
