#ifndef BatchLab_T_amanip_h
#define BatchLab_T_amanip_h

#include <AIDA/ITuple.h>
#include <AIDA/IEvaluator.h>
#include <AIDA/IFilter.h>

#include <Slash/Core/IWriter.h>

#include <Lib/Out.h>
#include <Lib/Cast.h>

#include <BatchLab/Core/BaseTuple.h> //FIXME

namespace BatchLab {

namespace T_amanip {

//////////////////////////////////////////////////////////////////////////////
template <class T>
bool project(
 Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,T& aFillable
,int aIndex
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  std::string type = aTuple.columnType(aIndex);
  if(type=="float") {
    float v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = aTuple.getFloat(aIndex);
      aFillable.fill((double)v,1);
    }
  } else if(type=="double") {
    double v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = aTuple.getDouble(aIndex);
      aFillable.fill(v,1);
    }
  } else if(type=="char") {
    char v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = aTuple.getChar(aIndex);
      aFillable.fill((double)v,1);
    }
  } else if(type=="byte") {
    BatchLab::BaseTuple* baseTuple = Lib_Cast(aTuple,BatchLab::BaseTuple);
    if(!baseTuple) {
      Lib::Out out(aPrinter);
      out << "BatchLab::T_amanip::project : "
          << " tuple not a BatchLab::BaseTuple."
          << Lib::endl;
      return false;
    }
    unsigned char v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = baseTuple->getByte(aIndex);
      aFillable.fill((double)v,1);
    }
  } else if(type=="short") {
    short v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = aTuple.getShort(aIndex);
      aFillable.fill((double)v,1);
    }
  } else if(type=="int") {
    int v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = aTuple.getInt(aIndex);
      aFillable.fill((double)v,1);
    }
  } else if(type=="long") {
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      aFillable.fill((double)aTuple.getLong(aIndex),1);
    }
  } else if(type=="boolean") {
    bool v;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      v = aTuple.getBoolean(aIndex);
      aFillable.fill((double)v,1);
    }
  } else {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " make no sense on column type " << type << "."
        << Lib::endl;
    return false;
  } 
  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T>
bool project(
 Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,T& aFillable
,AIDA::IEvaluator& aEvaluator
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aEvaluator.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  // Loop on entries :
  aTuple.start();
  while(aTuple.next()) {
    if(!aFilter.accept()) continue;
    aFillable.fill(aEvaluator.evaluateDouble(),1);
  }

  return true;
}

//////////////////////////////////////////////////////////////////////////////
template <class T>
bool project(
 Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,T& aFillable
,int aIndexX
,int aIndexY
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  std::string typex = aTuple.columnType(aIndexX);
  std::string typey = aTuple.columnType(aIndexY);
  if(typex!=typey) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " column type mismatch " << typex << " " << typey << "."
        << Lib::endl;
    return false;
  }
  if(typex=="float") {
    float vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getFloat(aIndexX);
      vy = aTuple.getFloat(aIndexY);
      aFillable.fill((double)vx,(double)vy,1);
    }
  } else if(typex=="double") {
    double vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getDouble(aIndexX);
      vy = aTuple.getDouble(aIndexY);
      aFillable.fill(vx,vy,1);
    }
  } else if(typex=="char") {
    char vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getChar(aIndexX);
      vy = aTuple.getChar(aIndexY);
      aFillable.fill((double)vx,(double)vy,1);
    }
  } else if(typex=="byte") {
    BatchLab::BaseTuple* baseTuple = Lib_Cast(aTuple,BatchLab::BaseTuple);
    if(!baseTuple) {
      Lib::Out out(aPrinter);
      out << "BatchLab::T_amanip::project : "
          << " tuple not a BatchLab::BaseTuple."
          << Lib::endl;
      return false;
    }
    unsigned char vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = baseTuple->getByte(aIndexX);
      vy = baseTuple->getByte(aIndexY);
      aFillable.fill((double)vx,(double)vy,1);
    }
  } else if(typex=="short") {
    short vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getShort(aIndexX);
      vy = aTuple.getShort(aIndexY);
      aFillable.fill((double)vx,(double)vy,1);
    }
  } else if(typex=="int") {
    int vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getInt(aIndexX);
      vy = aTuple.getInt(aIndexY);
      aFillable.fill((double)vx,(double)vy,1);
    }
  } else if(typex=="long") {
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      aFillable.fill((double)aTuple.getLong(aIndexX),
                     (double)aTuple.getLong(aIndexY),1);
    }
  } else if(typex=="boolean") {
    bool vx,vy;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getBoolean(aIndexX);
      vy = aTuple.getBoolean(aIndexY);
      aFillable.fill((double)vx,(double)vy,1);
    }
  } else {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " make no sense on column type " << typex << "."
        << Lib::endl;
    return false;
  } 
  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T>
bool project(
 Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,T& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aEvaluatorX.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  if(!aEvaluatorY.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  // Loop on entries :
  aTuple.start();
  while(aTuple.next()) {
    if(!aFilter.accept()) continue;
    aFillable.fill(aEvaluatorX.evaluateDouble(),
                   aEvaluatorY.evaluateDouble(),
                   1);
  }

  return true;
}


//////////////////////////////////////////////////////////////////////////////
template <class T>
bool project(
 Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,T& aFillable
,int aIndexX
,int aIndexY
,int aIndexZ
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  std::string typex = aTuple.columnType(aIndexX);
  std::string typey = aTuple.columnType(aIndexY);
  std::string typez = aTuple.columnType(aIndexZ);
  if((typex!=typey)||(typey!=typez)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " column type mismatch " 
        << typex << " " 
        << typey << " "
        << typez << "."
        << Lib::endl;
    return false;
  }
  if(typex=="float") {
    float vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getFloat(aIndexX);
      vy = aTuple.getFloat(aIndexY);
      vz = aTuple.getFloat(aIndexZ);
      aFillable.fill((double)vx,(double)vy,(double)vz,1);
    }
  } else if(typex=="double") {
    double vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getDouble(aIndexX);
      vy = aTuple.getDouble(aIndexY);
      vz = aTuple.getDouble(aIndexZ);
      aFillable.fill(vx,vy,vz,1);
    }
  } else if(typex=="char") {
    char vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getChar(aIndexX);
      vy = aTuple.getChar(aIndexY);
      vz = aTuple.getChar(aIndexZ);
      aFillable.fill((double)vx,(double)vy,(double)vz,1);
    }
  } else if(typex=="byte") {
    BatchLab::BaseTuple* baseTuple = Lib_Cast(aTuple,BatchLab::BaseTuple);
    if(!baseTuple) {
      Lib::Out out(aPrinter);
      out << "BatchLab::T_amanip::project : "
          << " tuple not a BatchLab::BaseTuple."
          << Lib::endl;
      return false;
    }
    unsigned char vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = baseTuple->getByte(aIndexX);
      vy = baseTuple->getByte(aIndexY);
      vz = baseTuple->getByte(aIndexZ);
      aFillable.fill((double)vx,(double)vy,(double)vz,1);
    }
  } else if(typex=="short") {
    short vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getShort(aIndexX);
      vy = aTuple.getShort(aIndexY);
      vz = aTuple.getShort(aIndexZ);
      aFillable.fill((double)vx,(double)vy,(double)vz,1);
    }
  } else if(typex=="int") {
    int vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getInt(aIndexX);
      vy = aTuple.getInt(aIndexY);
      vz = aTuple.getInt(aIndexZ);
      aFillable.fill((double)vx,(double)vy,(double)vz,1);
    }
  } else if(typex=="long") {
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      aFillable.fill((double)aTuple.getLong(aIndexX),
                     (double)aTuple.getLong(aIndexY),
                     (double)aTuple.getLong(aIndexZ),
                     1);
    }
  } else if(typex=="boolean") {
    bool vx,vy,vz;
    aTuple.start();
    while(aTuple.next()) {
      if(!aFilter.accept()) continue;
      vx = aTuple.getBoolean(aIndexX);
      vy = aTuple.getBoolean(aIndexY);
      vz = aTuple.getBoolean(aIndexZ);
      aFillable.fill((double)vx,(double)vy,(double)vz,1);
    }
  } else {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " make no sense on column type " << typex << "."
        << Lib::endl;
    return false;
  } 
  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T>
bool project(
 Slash::Core::IWriter& aPrinter
,AIDA::ITuple& aTuple
,T& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aEvaluatorX.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  if(!aEvaluatorY.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  if(!aEvaluatorZ.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Evaluator initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  if(!aFilter.initialize(aTuple)) {
    Lib::Out out(aPrinter);
    out << "BatchLab::T_amanip::project : "
        << " Filter initialization failed for tuple " 
        << Lib::sout(aTuple.title())
        << Lib::endl;
    return false;
  }

  // Loop on entries :
  aTuple.start();
  while(aTuple.next()) {
    if(!aFilter.accept()) continue;
    aFillable.fill(aEvaluatorX.evaluateDouble(),
                   aEvaluatorY.evaluateDouble(),
                   aEvaluatorZ.evaluateDouble(),
                   1);
  }

  return true;
}

}
              
}

#endif
